// ---------------------------------------------------------------------------
//	PSG Sound Implementation
//	Copyright (C) cisc 1997, 1999.
// ---------------------------------------------------------------------------
//	$Id: psg.cpp,v 1.2 1999/03/21 01:41:05 cisc Exp $

#include "headers.h"
#include "misc.h"
#include "psg.h"

uint8	PSG::noisetable[0x8000] = { 0, };	// mɍČꍇ 128K Kv
int		PSG::EmitTable[0x20] = { -1, };
uint	PSG::enveloptable[16][64] = { 0, };

// ---------------------------------------------------------------------------
//	mCYe[u쐬
//
void PSG::MakeNoiseTable()
{
	if (!noisetable[0])
	{
		int noise = 14321;
		for (int i=0; i<0x8000; i++)
		{
			noisetable[i] = uint8(noise & 1);
			noise = (noise >> 1) | (((noise << 14) ^ (noise << 16)) & 0x10000);
		}
	}
}

// ---------------------------------------------------------------------------
//	o̓e[u쐬
//	fɃe[uŎقȃXy[XB
//
void PSG::SetVolume(int volume)
{
	double base = 0x4000 / 3.0 * pow(10.0, volume / 40.0);
	for (int i=31; i>=2; i--)
	{
		EmitTable[i] = int(base);
		base /= 1.189207115;
	}
	EmitTable[1] = 0;
	EmitTable[0] = 0;
	MakeEnvelopTable();

	olevel[0] = EmitTable[(reg[ 8] & 15) * 2 + 1];
	olevel[1] = EmitTable[(reg[ 9] & 15) * 2 + 1];
	olevel[2] = EmitTable[(reg[10] & 15) * 2 + 1];
}

// ---------------------------------------------------------------------------
//	PCM f[^fo
//	dest		PCM f[^WJ|C^
//	nsamples	WJ PCM ̃Tv
//
#if 0
void PSG::EmitWave(Sample* dest, int nsamples)
{
}
#endif

// ---------------------------------------------------------------------------
//	Gx[vg`e[u
//
void PSG::MakeEnvelopTable()
{
	// 0 lo  1 up 2 down 3 hi
	static uint8 table1[16*2] =
	{
		2,0, 2,0, 2,0, 2,0, 1,0, 1,0, 1,0, 1,0,
		2,2, 2,0, 2,1, 2,3, 1,1, 1,3, 1,2, 1,0,
	};
	static uint8 table2[4] = {  0,  0, 31, 31 };
	static uint8 table3[4] = {  0,  1, -1,  0 };

	if (!enveloptable[0][0])
	{
		uint* ptr = enveloptable[0];

		for (int i=0; i<16*2; i++)
		{
			uint8 v = table2[table1[i]];
			
			for (int j=0; j<32; j++)
			{
				*ptr++ = EmitTable[v];
				v += table3[table1[i]];
			}
		}
	}
}

// ---------------------------------------------------------------------------
//	PSG ̃WX^ɒlZbg
//	regnum		WX^̔ԍ (0 - 15)
//	data		Zbgl
//
void PSG::SetReg(uint regnum, uint8 data)
{
	if (regnum < 0x10)
	{
		reg[regnum] = data;
		switch (regnum)
		{
			int tmp;

		case 0:		// ChA Fine Tune
		case 1:		// ChA Coarse Tune
			tmp = ((reg[0] + reg[1] * 256) & 0xfff);
			speriod[0] = tmp ? tperiodbase / tmp : tperiodbase;
			break;
		
		case 2:		// ChB Fine Tune
		case 3:		// ChB Coarse Tune
			tmp = ((reg[2] + reg[3] * 256) & 0xfff);
			speriod[1] = tmp ? tperiodbase / tmp : tperiodbase;	  
			break;
		
		case 4:		// ChC Fine Tune
		case 5:		// ChC Coarse Tune
			tmp = ((reg[4] + reg[5] * 256) & 0xfff);
			speriod[2] = tmp ? tperiodbase / tmp : tperiodbase;	  
			break;

		case 6:		// Noise generator control
			data &= 0x1f;
			nperiod = data ? nperiodbase / data : nperiodbase;
			break;

		case 8:
			olevel[0] = EmitTable[(data & 15) * 2 + 1];
			break;

		case 9:
			olevel[1] = EmitTable[(data & 15) * 2 + 1];
			break;
		
		case 10:
			olevel[2] = EmitTable[(data & 15) * 2 + 1];
			break;

		case 11:	// Envelop period
		case 12:
			tmp = ((reg[11] + reg[12] * 256) & 0xffff);
			eperiod = tmp ? eperiodbase / tmp : eperiodbase * 2;
			break;

		case 13:	// Envelop shape
			ecount = 0;
			envelop = enveloptable[data & 15];
			break;
		}
	}
}

// ---------------------------------------------------------------------------
//	PSG (RESET M) 
//
void PSG::Reset()
{
	for (int i=0; i<14; i++)
		SetReg(i, 0);
	SetReg(7, 0xff);
	SetReg(14, 0xff);
	SetReg(15, 0xff);
}

// ---------------------------------------------------------------------------
//	NbNg̐ݒ
//
void PSG::SetPSGClock(int clock, int rate)
{
	tperiodbase = int((1 << SSG_TONESHIFT ) / 4.0 * clock / rate);
	eperiodbase = int((1 << SSG_ENVSHIFT  ) / 4.0 * clock / rate);
	nperiodbase = int((1 << SSG_NOISESHIFT) / 4.0 * clock / rate);
	
	// ef[^̍XV
	int tmp;
	tmp = ((reg[0] + reg[1] * 256) & 0xfff);
	speriod[0] = tmp ? tperiodbase / tmp : tperiodbase;
	tmp = ((reg[2] + reg[3] * 256) & 0xfff);
	speriod[1] = tmp ? tperiodbase / tmp : tperiodbase;
	tmp = ((reg[4] + reg[5] * 256) & 0xfff);
	speriod[2] = tmp ? tperiodbase / tmp : tperiodbase;
	tmp = reg[6] & 0x1f;
	nperiod = tmp ? nperiodbase / tmp / 2 : nperiodbase / 2;
	tmp = ((reg[11] + reg[12] * 256) & 0xffff);
	eperiod = tmp ? eperiodbase / tmp : eperiodbase * 2;
}

// ---------------------------------------------------------------------------
//	RXgN^EfXgN^
//
PSG::PSG()
{
	SetVolume(0);
	MakeNoiseTable();
	Reset();
}

PSG::~PSG()
{

}
