// ---------------------------------------------------------------------------
//	M88 - PC-8801 Emulator
//	Copyright (C) cisc 1997, 1999.
// ---------------------------------------------------------------------------
//	$Id: sound.cpp,v 1.7 1999/04/07 13:44:27 cisc Exp $

#include "headers.h"
//#include <string.h>
//#include <math.h>
#include "types.h"
#include "misc.h"
#include "schedule.h"
#include "pc88/sound.h"
#include "pc88/pc88.h"
#include "pc88/config.h"
#include "fmgen.h"

#define LOGNAME "sound"
#include "diag.h"

using namespace PC8801;

// ---------------------------------------------------------------------------
//	Ej
//
YM2608::YM2608(const ID& id, int i)
: Device(id)
{
	scheduler = 0;
	sound = 0;
	imaskbit = i;
	opnamode = newopnamode = false;
	nextcount = 0;
}

YM2608::~YM2608()
{
}

// ---------------------------------------------------------------------------
//	
//
bool YM2608::Init(Bus* b, int intrport, int io, Scheduler* s, uint c, uint r)
{
	bus = b;
	scheduler = s;
	portio = io;
	opn.SetIntr(bus, intrport);
	
	if (!psg.Init(r))
		return false;
	if (!opn.Init(c, r, 0))
		return false;
	if (!SetRate(r, c))
		return false;
	prevtime = scheduler->GetTime();
	TimeEvent();
	return true;
}

// ---------------------------------------------------------------------------
//	EĐ[gݒ
//
bool YM2608::SetRate(uint _clock, uint rate)
{
	clock = _clock;
	bool fm55k = false;
	srate = rate;
	if (rate == SOUND_55K)
	{
		fm55k = true;
		srate = 44100;
	}
	psg.Init(srate);
	opn.SetRate(clock, srate, fm55k);
	psg.SetPSGClock(clock/4, srate); 
	return true;
}

// ---------------------------------------------------------------------------
//	
//
void YM2608::Mix(FM::Sample* dest, int nsamples, int b1, int b2, int b3)
{
	opn.Mix(dest, nsamples);
	psg.EmitWaveDCh(dest, nsamples, b1, b2, b3);
}

// ---------------------------------------------------------------------------
//	ʐݒ
//
static inline int ConvertVolume(int volume)
{
	return volume > -40 ? volume : -200;
}

void YM2608::SetVolume(const Config* config)
{
	opn.SetVolumeFM(ConvertVolume(config->volfm));
	opn.SetVolumeADPCM(ConvertVolume(config->voladpcm));
	opn.SetVolumeRhythmTotal(ConvertVolume(config->volrhythm));
	opn.SetVolumeRhythm(0, ConvertVolume(config->volbd));
	opn.SetVolumeRhythm(1, ConvertVolume(config->volsd));
	opn.SetVolumeRhythm(2, ConvertVolume(config->voltop));
	opn.SetVolumeRhythm(3, ConvertVolume(config->volhh));
	opn.SetVolumeRhythm(4, ConvertVolume(config->voltom));
	opn.SetVolumeRhythm(5, ConvertVolume(config->volrim));
	
	psg.SetVolume(ConvertVolume(config->volssg));
}

// ---------------------------------------------------------------------------
//	Reset
//
void YM2608::Reset(uint, uint)
{
	opnamode = newopnamode;
	
	opn.Reset();
	psg.Reset();
	psg.SetPSGClock(clock/4, srate);
	opn.SetIntrMask(false);
}

// ---------------------------------------------------------------------------
//	荞
//
void YM2608::OPNUnit::Intr(bool flag)
{
	bool prev = intrpending && intrenabled && bus;
	intrpending = flag;
	if (intrpending && intrenabled && bus && !prev)
		bus->Out(pintr, true);
}

// ---------------------------------------------------------------------------
//	荞݋H
//
inline void YM2608::OPNUnit::SetIntrMask(bool en)
{
	bool prev = intrpending && intrenabled && bus;
	intrenabled = en;
	if (intrpending && intrenabled && bus && !prev)
		bus->Out(pintr, true);
}

void YM2608::SetIntrMask(uint, uint intrmask)
{
	opn.SetIntrMask(!(imaskbit & intrmask));
}

// ---------------------------------------------------------------------------
//	SetRegisterIndex
//
void YM2608::SetIndex0(uint, uint data)
{
	index0 = data;
	if (enable && (index0 & 0xfc) == 0x2c)
	{
		switch (index0 & 3)
		{
		case 0: return;
		case 1: psg.SetPSGClock(clock/4, srate); break;
		case 2: psg.SetPSGClock(clock/2, srate); break;
		case 3:	psg.SetPSGClock(clock/1, srate); break;
		}
		opn.SetReg(index0, 0);
	}
}

void YM2608::SetIndex1(uint, uint data)
{
	index1 = data;
}

// ---------------------------------------------------------------------------
//	WriteRegister
//
void YM2608::WriteData0(uint, uint data)
{
	if (enable)
	{
		if (index0 < 0x10)
			psg.SetReg(index0, data);
		else
		{
			if (!opnamode)
			{
				if ((index0 & 0xf0) == 0x10) return;
				if (index0 == 0x22) data = 0;
				if (index0 == 0x29) data = 3;
				if (index0 >= 0xb4) data = 0xc0;
			}
			opn.SetReg(index0, data);
			if (index0 == 0x27)
			{
				LOG0("Port27\t");
				UpdateTimer();
			}
		}
	}
}

void YM2608::WriteData1(uint, uint data)
{
	if (enable && opnamode)
		opn.SetReg(0x100 | index1, data);
}

// ---------------------------------------------------------------------------
//	ReadRegister
//
uint YM2608::ReadData0(uint)
{
	if (!enable)
		return 0xff;

	if ((index0 & 0xfe) == 0x0e)
		return bus->In(portio + (index0 & 1));
	if (index0 < 0x10)
		return psg.GetReg(index0);
	if (index0 == 0xff && !opnamode)
		return 0;
	return opn.GetReg(index0);
}

uint YM2608::ReadData1(uint)
{
	return enable && opnamode ? opn.GetReg(0x100 | index1) : 0xff;
}

// ---------------------------------------------------------------------------
//	ReadStatus
//
uint YM2608::ReadStatus(uint)
{
	return enable ? opn.ReadStatus() : 0xff;
}

uint YM2608::ReadStatusEx(uint)
{
	return enable && opnamode ? opn.ReadStatusEx() : 0xff;
}

// ---------------------------------------------------------------------------
//	SetOPNMode
//
void YM2608::SetOPNMode(bool opna)
{
	newopnamode = opna;
}

// ---------------------------------------------------------------------------
//	^C}[XV
//
void YM2608::UpdateTimer()
{
	scheduler->DelEvent(this);
	
	nextcount = opn.GetNextEvent();
	if (nextcount)
	{
		nextcount = (nextcount+9) / 10;
		scheduler->AddEvent(nextcount, this, static_cast<TimeFunc>(TimeEvent), 0);
//		LOG1("AddEvent: %d\n", nextcount);
	}
}

// ---------------------------------------------------------------------------
//	^C}[
//
void YM2608::TimeEvent(uint rtc)
{
	int currenttime = scheduler->GetTime();
	int diff = currenttime - prevtime;
	prevtime = scheduler->GetTime();
	
//	LOG3("\tTimeEvent : current:%d  diff:%d  intended count:%d\n", currenttime, diff, nextcount);
	
	if (opn.Count(diff * 10) || !rtc)
	{
//		LOG0("Time\t");
		UpdateTimer();
	}
	
	if (sound)
		sound->OPNEvent();
}


// ---------------------------------------------------------------------------
//	Ej
//
Sound::Sound(const ID& id, const ID& idopn1, const ID& idopn2)
: Device(id), opn1(idopn1, 0x80), opn2(idopn2, 0x80)
{
	uscount = 0;
	buffer = 0;
	bwpos = 0;
	enabled = false;
	opn2enabled = false;
	mixingbuf = 0;
	beepmask = 0xa0;

	pindex = 0;
	pswitches = 0;
	bindex = 0;
	bswitches = 0;
	bstep = 0;
	pstep = 0;
	port40 = 0;
	cfgflg = 0;
}

Sound::~Sound()
{
	Cleanup();
}

// ---------------------------------------------------------------------------
//	Ƃ
//	rate = SOUND_55K, 44100, 22050, 11025, 0 ̂ǂꂩ
//
bool Sound::Init(PC88* pc88, uint rate, uint32 opnclock, int bufsize)
{
	pc = pc88;

	prevtime = pc->GetTime();

	enabled = false;
	
	if (!opn1.Init(pc88->GetBus(), PC88::pint4, PC88::popnio, pc, opnclock, rate))
		return false;
	if (!opn2.Init(pc88->GetBus(), PC88::pint4, PC88::popnio, pc, opnclock, rate))
		return false;
	opn1.SetSound(this);
	opn2.SetSound(this);
	
	if (!SetRate(rate, opnclock, bufsize))
		return false;
	
	return true;
}

// ---------------------------------------------------------------------------
//	[gݒ
//	rate = SOUND_55K, 44100, 22050, 11025, 0 ̂ǂꂩ
//
bool Sound::SetRate(uint rate, uint32 clock, int bufsize)
{
	opn1.SetRate(clock, rate);
	opn2.SetRate(clock, rate);
	
	if (rate == SOUND_55K)
		rate = 44100;
	
	enabled = false;
	samplingrate = rate;
	
	if (!bufsize)
		return true;

	srate50 = rate / 50;
	buffersize = bufsize;
	bvalue = int(100000.0 * 65536.0 / rate);

	delete[] buffer; delete[] mixingbuf;

	buffer = new uint8[buffersize];
	mixingbuf = new int32[buffersize/2];
	if (!buffer || !mixingbuf)
		return false;

	memset(buffer, 0, buffersize);

	read = 0; write = 0;
	enabled = true;

	return true;
}


// ---------------------------------------------------------------------------
//	Еt
//
void Sound::Cleanup()
{
	delete[] buffer; delete[] mixingbuf;
	buffer = 0; mixingbuf = 0;
}

// ---------------------------------------------------------------------------
//	Reset
//
void Sound::Reset(uint, uint)
{
	opn2enabled = (cfgflg & (Config::opnaona8 | Config::opnona8)) != 0;

	opn1.Reset();
	opn1.Enable(true);
	opn2.Reset();
	opn2.Enable(opn2enabled);
}

// ---------------------------------------------------------------------------
//	ԏ
//
void Sound::OPNEvent()
{
	uint32 currenttime = pc->GetTime();
	uint32 uscount = currenttime - prevtime;
	prevtime = currenttime;
//	LOG0("OPNEV\n");
//	LOG1("ev:%d\n", uscount);
	
	int nsamples = ((uscount * srate50 / 1000) & ~1) * 2;
	LOG1("Store = %5d bytes\n", nsamples);
	Store(nsamples);
}

void Sound::RTC(uint,uint)
{
	int d;
	UpdateBeepState();
	if ((d = pc->GetTime() - prevtime) > 500)	// 5ms
	{
//		LOG1("R(%d)", d);
		opn1.TimeEvent(1);
//		opn2.TimeEvent();
	}
}

// ---------------------------------------------------------------------------
//	Oobt@ɉǉ
//
void Sound::Store(int bytes)
{
	if (!enabled)
		return;

	// Oobt@̋󂫂vZ
	int free;
	if (read <= write)
		free = buffersize + read - write;
	else
		free = read - write;

	// ނׂf[^ʂvZ
//	if (free-4 < bytes)
//		LOG1("xfill = %5d bytes\n", bytes - free - 4);
	bytes = Min(bytes, free-4);
	if (bytes < 0)
		return;
	
//	LOG4("read = %6d  write = %6d  free = %6d  bytes = %6d.\n", read, write, free, bytes);

	// 
	if (buffersize - write >= bytes)
	{
		// xŏꍇ
		Mix(&buffer[write], bytes);
	}
	else
	{
		// Qxɕďꍇ
		Mix(&buffer[write], buffersize - write);
		Mix(&buffer[0], bytes - (buffersize - write));
	}
	write += bytes;
	if (write >= buffersize)
		write -= buffersize;
}

// ---------------------------------------------------------------------------
//	Oobt@特Ⴄ
//
void Sound::Fill(void* d, int bytes)
{
	if (!enabled)
		return;

//	LOG1("Fill  = %5d bytes\n", bytes);
	uint8* dest = (uint8*) d;
	while (bytes > 0)
	{
		int xsize = Min(bytes, buffersize - 4);
		xsize = Min(xsize, buffersize - read);
		
		int avail;
		if (write >= read)
			avail = write - read;
		else
			avail = buffersize + write - read;

		if (xsize > avail)
		{
//			LOG1("xstore= %5d bytes\n", xsize-avail);
			Store(xsize - avail);
			uscount = 0;
		}

		memcpy(dest, buffer + read, xsize);
		dest += xsize;
		bytes -= xsize;
		read += xsize;
		if (read >= buffersize)
			read -= buffersize;
	}
}

// ---------------------------------------------------------------------------
//	
//
void Sound::Mix(void* d, int bytes)
{
	int nsamples = bytes / 4;
	memset(mixingbuf, 0, nsamples * 8);
	opn1.Mix(mixingbuf, nsamples, bstep, pstep, port40);
	if (opn2enabled)
		opn2.Mix(mixingbuf, nsamples);

	int32* src = mixingbuf;
	int16* dest = (int16*) d;
	for (; nsamples>0; nsamples--)
	{
		*dest++ = Limit(*src++, 32767, -32768);
		*dest++ = Limit(*src++, 32767, -32768);
	}
}

// ---------------------------------------------------------------------------
//	ʐݒ
//
void Sound::SetVolume(const Config* config)
{
	opn1.SetVolume(config);
	opn2.SetVolume(config);
}

// ---------------------------------------------------------------------------
//	ݒXV
//
void Sound::ApplyConfig(const Config* config)
{
	cfgflg = config->flags;
	opn1.SetOPNMode((config->flags & Config::enableopna) != 0);
	opn2.SetOPNMode((config->flags & Config::opnaona8) != 0);
	SetVolume(config);

	beepmask = (config->flags & Config::disablesing) ? 0x20 : 0xa0;
	port40 &= beepmask;
}

// ---------------------------------------------------------------------------
//	BEEP Port ւ Out
//
void Sound::Out40(uint, uint data)
{
	data &= beepmask;
	uint time = pc->GetTime();		// 10 us Pʂ̎
	uint dtime = plasttime-time;
	plasttime = time;
	
	uint modified = data ^ port40;
	port40 = data;

	if (modified & 0x80)
	{
		pcount[(pindex++) & 1] = Min(0x1000000, time - plasttime);
		plasttime = time;
		pswitches = 10;
	}
	if (modified & 0x20)
	{
		LOG1("time = %d\n", time-blasttime);
		bcount[(bindex++) & 1] = Min(0x1000000, time - blasttime);
		blasttime = time;
		bswitches = 10;
	}
}

// ---------------------------------------------------------------------------
//	BEEP pIXV (600Hz)
//
void Sound::UpdateBeepState()
{
	uint level;
	if (pswitches > 0)
	{
		level = (pcount[0] + pcount[1]) / 2 + 1;
		pstep = bvalue / level;
		pswitches--;
	}
	else
		pstep = 0;
	
	if (bswitches > 0)
	{
		level = (bcount[0] + bcount[1]) / 2 + 1;
		bstep = bvalue / level;
		bswitches--;
	}
	else
		bstep = 0;
}


// ---------------------------------------------------------------------------
//	obt@󂩁Cɋ߂Ԃ?
//
bool Sound::IsBufferEmpty()
{
	int avail;
	if (write >= read)
		avail = write - read;
	else
		avail = buffersize + write - read;
	
	return avail == 0;
}

// ---------------------------------------------------------------------------
//	device description
//
const Device::Descriptor YM2608::descriptor = { indef, outdef };

const Device::OutFuncPtr YM2608::outdef[] = 
{
	static_cast<OutFuncPtr> (Reset),
	static_cast<OutFuncPtr> (SetIndex0),
	static_cast<OutFuncPtr> (SetIndex1),
	static_cast<OutFuncPtr> (WriteData0),
	static_cast<OutFuncPtr> (WriteData1),
	static_cast<OutFuncPtr> (SetIntrMask),
};

const Device::InFuncPtr YM2608::indef[] = 
{
	static_cast<InFuncPtr> (ReadStatus),
	static_cast<InFuncPtr> (ReadStatusEx),
	static_cast<InFuncPtr> (ReadData0),
	static_cast<InFuncPtr> (ReadData1),
};

const Device::Descriptor Sound::descriptor = { 0, outdef };

const Device::OutFuncPtr Sound::outdef[] = 
{
	static_cast<OutFuncPtr> (Reset),
	static_cast<OutFuncPtr> (RTC),
	static_cast<OutFuncPtr> (Out40),
};

//const Device::InFuncPtr Sound::indef[] = 
//{
//};

